@@ -15,10 +15,12 @@
 """
 from __future__ import print_function
 
-import subprocess
 import errno
 import gc
+import io
+import os
 import re
+import subprocess
 import sys
 import threading
 import time
@@ -592,6 +594,25 @@ def spawn_layer_in_subprocess(result, script_parts, options, features,
             child.communicate()
 
 
+def _get_output_buffer(stream):
+    """Get a binary-safe version of a stream."""
+    try:
+        fileno = stream.fileno()
+    except (io.UnsupportedOperation, AttributeError):
+        pass
+    else:
+        # Win32 mangles \r\n to \n and that breaks streams.  See
+        # https://bugs.launchpad.net/bugs/505078.
+        if sys.platform == 'win32':
+            import msvcrt
+            msvcrt.setmode(fileno, os.O_BINARY)
+    try:
+        stream.write(b'')
+    except TypeError:
+        return stream.buffer
+    return stream
+
+
 class AbstractSubprocessResult(object):
     """A result of a subprocess layer run."""
 
@@ -611,8 +632,6 @@ class DeferredSubprocessResult(AbstractSubprocessResult):
     """Keeps stdout around for later processing,"""
 
     def write(self, out):
-        if not isinstance(out, str): # It's binary, which means it's Python 3
-            out = out.decode()
         if not _is_dots(out):
             self.stdout.append(out)
 
@@ -620,18 +639,17 @@ class DeferredSubprocessResult(AbstractSubprocessResult):
 class ImmediateSubprocessResult(AbstractSubprocessResult):
     """Sends complete output to queue."""
 
+    def __init__(self, layer_name, queue):
+        super(ImmediateSubprocessResult, self).__init__(layer_name, queue)
+        self.stream = _get_output_buffer(sys.stdout)
+
     def write(self, out):
-        if not isinstance(out, str):
-            # In Python 3, a Popen process stdout uses bytes,
-            # While normal stdout uses strings. So we need to convert
-            # from bytes to strings here.
-            out = out.decode()
-        sys.stdout.write(out)
+        self.stream.write(out)
         # Help keep-alive monitors (human or automated) keep up-to-date.
-        sys.stdout.flush()
+        self.stream.flush()
 
 
-_is_dots = re.compile(r'\.+(\r\n?|\n)').match # Windows sneaks in a \r\n.
+_is_dots = re.compile(br'\.+(\r\n?|\n)').match # Windows sneaks in a \r\n.
 class KeepaliveSubprocessResult(AbstractSubprocessResult):
     "Keeps stdout for later processing; sends marks to queue to show activity."
 
@@ -644,8 +662,6 @@ class KeepaliveSubprocessResult(AbstractSubprocessResult):
     done = property(lambda self: self._done, _set_done)
 
     def write(self, out):
-        if not isinstance(out, str): # It's binary, which means it's Python 3
-            out = out.decode()
         if _is_dots(out):
             self.queue.put((self.layer_name, out.strip()))
         else:
@@ -658,7 +674,8 @@ def resume_tests(script_parts, options, features, layers, failures, errors,
     stdout_queue = None
     if options.processes == 1:
         result_factory = ImmediateSubprocessResult
-    elif options.verbose > 1:
+    elif (options.verbose > 1 and
+            not options.subunit and not options.subunit_v2):
         result_factory = KeepaliveSubprocessResult
         stdout_queue = Queue.Queue()
     else:
@@ -680,6 +697,7 @@ def resume_tests(script_parts, options, features, layers, failures, errors,
     current_result = next(results_iter)
     last_layer_intermediate_output = None
     output = None
+    stdout = _get_output_buffer(sys.stdout)
     while ready_threads or running_threads:
         while len(running_threads) < options.processes and ready_threads:
             thread = ready_threads.pop(0)
@@ -687,7 +705,7 @@ def resume_tests(script_parts, options, features, layers, failures, errors,
             running_threads.append(thread)
 
         for index, thread in reversed(list(enumerate(running_threads))):
-            if not thread.isAlive():
+            if not thread.is_alive():
                 del running_threads[index]
 
         # Clear out any messages in queue
@@ -700,18 +718,19 @@ def resume_tests(script_parts, options, features, layers, failures, errors,
             if layer_name != last_layer_intermediate_output:
                 # Clarify what layer is reporting activity.
                 if previous_output is not None:
-                    sys.stdout.write(']\n')
-                sys.stdout.write(
-                    '[Parallel tests running in %s:\n  ' % (layer_name,))
+                    stdout.write(b']\n')
+                stdout.write(
+                    ('[Parallel tests running in '
+                     '%s:\n  ' % (layer_name,)).encode('utf-8'))
                 last_layer_intermediate_output = layer_name
-            sys.stdout.write(output)
+            stdout.write(output)
         # Display results in the order they would have been displayed, had the
         # work not been done in parallel.
         while current_result and current_result.done:
             if output is not None:
-                sys.stdout.write(']\n')
+                stdout.write(b']\n')
                 output = None
-            sys.stdout.writelines(current_result.stdout)
+            stdout.writelines(current_result.stdout)
 
             try:
                 current_result = next(results_iter)
@@ -719,7 +738,7 @@ def resume_tests(script_parts, options, features, layers, failures, errors,
                 current_result = None
 
         # Help keep-alive monitors (human or automated) keep up-to-date.
-        sys.stdout.flush()
+        stdout.flush()
         time.sleep(0.01) # Keep the loop from being too tight.
 
     # Return the total number of tests run.
@@ -908,7 +927,7 @@ class TestResult(unittest.TestResult):
         # Did the test leave any new threads behind?
         new_threads = []
         for t in threading.enumerate():
-            if t.isAlive() and t not in self._threads:
+            if t.is_alive() and t not in self._threads:
                 if not any([re.match(p, t.name)
                             for p in self.options.ignore_new_threads]):
                     new_threads.append(t)
@@ -957,7 +976,7 @@ def layer_sort_key(layer):
         # We make the simplifying assumption that the order of initialization
         # of base layers does not matter.  Given that, traversing the bases
         # in reverse order here keeps the ordering of layers in
-        # testrunner-layers.txt the same as it was in older versions of
+        # testrunner-layers.rst the same as it was in older versions of
         # zope.testrunner, so let's use that.
         for base in layer.__bases__[::-1]:
             if base is not object and base not in seen:
